/**
 * @module xethya-extension-range
 * @namespace Xethya.Utils
 */

import { Assertable, assert } from 'xethya-extension-base';

/**
 * Represents a range of number, described through boundaries.
 *
 * @property {Number} lowerBound - The minimum number in this range.
 * @property {Number} upperBound - The maximum number in this range.
 */

@Assertable
export class Range {
  /**
   * Instantiates a Range object.
   * 
   * @constructor
   * @param  {Number} lowerBound - The minimum number in this range.
   * @param  {Number} upperBound - The maximum number in this range.
   * @throws {Error} If lowerBound and upperBound are the same.
   */
  constructor(lowerBound, upperBound) {
    this.assert(lowerBound !== undefined, 'Range#constructor: lowerBound is required');
    this.assert(typeof lowerBound === 'number', 'Range#constructor: lowerBound must be a Number');
    this.assert(upperBound !== undefined, 'Range#constructor: upperBound is required');
    this.assert(typeof upperBound === 'number', 'Range#constructor: upperBound must be a Number');
    this.assert(lowerBound !== upperBound, 'Range#constructor: lowerBound and upperBound cannot be equal');

    this.lowerBound = Math.min(lowerBound, upperBound);
    this.upperBound = Math.max(lowerBound, upperBound);
  }

  /**
   * Checks if a value is in the defined range.
   *
   * @public
   * @function
   * @param  {Number} value - Value to compare.
   * @return {Boolean} true if in range, false otherwise.
   */
  includes(value) {
    this.assert(value !== undefined, 'Range#includes: value is required');
    this.assert(typeof value === 'number', 'Range#includes: value must be a Number');

    return this.lowerBound <= value && value <= this.upperBound;
  }

  /**
   * Converts the Range object to a string representation.
   * 
   * @public
   * @function
   * @return {String}
   */
  toString() {
    return `${this.lowerBound.toString()} ~ ${this.upperBound.toString()}`;
  }

  /**
   * Creates a Range from an array of two numbers.
   *
   * @public
   * @static
   * @function
   * @param  {Array.<Number>} values - Boundaries of the range.
   * @return {Range}
   */
  static fromArray(values) {
    const errorMessage = 'Range#fromArray: values must be an Array of 2 numerical elements';

    assert(Array.isArray(values), errorMessage);
    assert(values.length === 2, errorMessage);
    assert(values.every(value => typeof value === 'number'), errorMessage);

    return new Range(Number(values[0]), Number(values[1]));
  }

  /**
   * Creates a Range from a string-based notation.
   *
   * @public
   * @static
   * @function
   * @param  {String} notedRange - A string representation of a Range,
   *                  using delimiters. Accepted formats: x,y x;y x:y x~y.
   * @return {Range}
   */
  static fromNotation(notedRange) {
    const errorMessage = 'Range#fromNotation: notedRange must use one of these formats: x,y x;y x:y x~y';

    assert(notedRange !== undefined, errorMessage);
    assert(typeof notedRange === 'string', errorMessage);

    let range;
    const allowedDelimiters = [',', ';', ':', '~'];

    assert(allowedDelimiters.some(delimiter => notedRange.includes(delimiter)), errorMessage);

    let delimiterFound = false;

    while (!delimiterFound) {
      const delimiter = allowedDelimiters.shift();

      delimiterFound = notedRange.includes(delimiter);

      if (delimiterFound) {
        const data = notedRange.split(delimiter).map(d => d.trim());

        assert(data.length === 2, errorMessage);

        range = Range.fromArray(data.map(d => Number(d)));
      }
    }

    assert(range !== undefined, errorMessage);

    return range;
  }
}
